import Lain from "../Lain.js";
import Library from "../Library.js";
import NodeSurface from "./NodeSurface.js";
import Source from "./Source.js";
import Path from 'path'
import FS from 'fs/promises'
import { Image, Path2D } from "@napi-rs/canvas";

// Polyfill
global.Image = Image
global.Path2D = Path2D

function Client(directory) {
  this.workingDirectory = directory
  this.surface = new NodeSurface(this)
  this.library = new Library(this)
  this.lain = new Lain(this.library)
  this.source = new Source(this)

  this.run = (args, program) => {
    if(program.indexOf('$') > -1) {
      program = this.replaceConstants(program, args)
    }

    this.surface.clear()
    return this.lain.run('(' + program + ')')
  }

  this.bind = () => {
    // Perhaps binding in the cli could work by piping in 
    // CSV. Events could be recorded in the web client and saved
    // to the format
  }

  this.log = (...msg) => {
    console.log('Evaluation:', ...msg)
  }

  this.cache = {
    data: {},
    set: (key, content) => {
      this.log((this.cache.data[key] ? 'Updated ' : 'Stored ') + key)
      this.cache.data[key] = content
    },

    get: (key) => {
      return this.cache.data[key]
    },

    // Async
    load: (pathResolvable) => {
      let key = resolvePath(pathResolvable)

      return openImage(key)
    },

    // Async
    resolve: (pathResolvable) => {
      let key = resolvePath(pathResolvable),
          out = this.cache.get(key)

      if(out) {
        return Promise.resolve(out)
      } else {
        return openImage(key)
      }
    }
  }

  const replaceConstants = (txt, args) => {
    return txt.replaceAll(/\$(\d+)/g, (match, num) => {
      return JSON.stringify(args[parseInt(num)])
    })
  }

  const resolvePath = pathResolvable => {
    return Path.isAbsolute(pathResolvable) ? Path.normalize(pathResolvable) : Path.join(this.workingDirectory, pathResolvable)
  }

  const openImage = async (key) => {
    let buf = await FS.readFile(key)
      .catch(err => {
        console.error('Source', `Could not load image at path "${key}"`, err)
        process.exit()
      })

    let img = new Image()
    img.src = buf
    this.cache.set(key, buf)
    return buf
  }
}

export default Client